@@ -85,6 +85,28 @@ class JAXTrainer(base_trainer.Trainer):
             metrics_variables,
         )
 
+    def _update_metrics_variables(
+        self, metrics_variables, unscaled_loss, x, y, y_pred, sample_weight
+    ):
+        with backend.StatelessScope(
+            state_mapping=[
+                (ref_v, v)
+                for ref_v, v in zip(self.metrics_variables, metrics_variables)
+            ]
+        ) as scope:
+            self._loss_tracker.update_state(
+                unscaled_loss, sample_weight=tree.flatten(x)[0].shape[0]
+            )
+            logs = self.compute_metrics(x, y, y_pred, sample_weight)
+
+        new_metrics_variables = []
+        for ref_v in self.metrics_variables:
+            new_v = scope.get_current_value(ref_v)
+            if new_v is None:
+                new_v = ref_v.value
+            new_metrics_variables.append(new_v)
+        return logs, new_metrics_variables
+
     def train_step(self, state, data):
         (
             trainable_variables,
@@ -117,24 +139,9 @@ class JAXTrainer(base_trainer.Trainer):
             optimizer_variables, grads, trainable_variables
         )
 
-        with backend.StatelessScope(
-            state_mapping=[
-                (ref_v, v)
-                for ref_v, v in zip(self.metrics_variables, metrics_variables)
-            ]
-        ) as scope:
-            self._loss_tracker.update_state(
-                unscaled_loss, sample_weight=tree.flatten(x)[0].shape[0]
-            )
-            logs = self.compute_metrics(x, y, y_pred, sample_weight)
-
-        new_metrics_variables = []
-        for ref_v in self.metrics_variables:
-            new_v = scope.get_current_value(ref_v)
-            if new_v is None:
-                new_v = ref_v.value
-            new_metrics_variables.append(new_v)
-        metrics_variables = new_metrics_variables
+        logs, metrics_variables = self._update_metrics_variables(
+            metrics_variables, unscaled_loss, x, y, y_pred, sample_weight
+        )
 
         state = self._enforce_jax_state_sharding(
             trainable_variables,
@@ -164,24 +171,9 @@ class JAXTrainer(base_trainer.Trainer):
             aux
         )
 
-        with backend.StatelessScope(
-            state_mapping=[
-                (ref_v, v)
-                for ref_v, v in zip(self.metrics_variables, metrics_variables)
-            ]
-        ) as scope:
-            self._loss_tracker.update_state(
-                unscaled_loss, sample_weight=tree.flatten(x)[0].shape[0]
-            )
-            logs = self.compute_metrics(x, y, y_pred, sample_weight)
-
-        new_metrics_variables = []
-        for ref_v in self.metrics_variables:
-            new_v = scope.get_current_value(ref_v)
-            if new_v is None:
-                new_v = ref_v.value
-            new_metrics_variables.append(new_v)
-        metrics_variables = new_metrics_variables
+        logs, metrics_variables = self._update_metrics_variables(
+            metrics_variables, unscaled_loss, x, y, y_pred, sample_weight
+        )
 
         (
             trainable_variables,
@@ -224,108 +216,109 @@ class JAXTrainer(base_trainer.Trainer):
         )
         return outputs, non_trainable_variables
 
+    def _make_function(self, step_function, concatenate_outputs=False):
+        if self.steps_per_execution > 1:
+            if concatenate_outputs:
+
+                def concatenate(outputs):
+                    output = outputs[0]
+                    for next_output in outputs[1:]:
+                        output = tree.map_structure(
+                            lambda t1, t2: jax.numpy.concatenate([t1, t2]),
+                            output,
+                            next_output,
+                        )
+                    return output
+
+                if not self.run_eagerly and self.jit_compile:
+                    concatenate = jax.jit(concatenate)
+
+                def iterator_step(state, iterator):
+                    data = next(iterator)
+                    outputs, state = step_function(state, data)
+                    outputs = [outputs]
+                    try:
+                        for _ in range(self.steps_per_execution - 1):
+                            data = next(iterator)
+                            _outputs, state = step_function(state, data)
+                            outputs.append(_outputs)
+                    except StopIteration:
+                        pass
+                    outputs = concatenate(outputs)
+                    return outputs, state
+
+            else:
+
+                def iterator_step(state, iterator):
+                    data = next(iterator)
+                    outputs, state = step_function(state, data)
+                    try:
+                        for _ in range(self.steps_per_execution - 1):
+                            data = next(iterator)
+                            outputs, state = step_function(state, data)
+                    except StopIteration:
+                        pass
+                    return outputs, state
+
+        else:
+
+            def iterator_step(state, iterator):
+                return step_function(state, next(iterator))
+
+        return iterator_step
+
     def make_train_function(self, force=False):
         if self.train_function is not None and not force:
             return
-
-        def one_train_step(state, data):
-            data = data[0]
-            return self.train_step(state, data)
-
-        def multi_train_steps(state, data):
-            for single_step_data in data:
-                logs, state = one_train_step(state, [single_step_data])
-            return logs, state
-
-        if self.steps_per_execution > 1:
-            train_step = multi_train_steps
-        else:
-            train_step = one_train_step
-
         if not self.run_eagerly and self.jit_compile:
-            # Note that we mark the state and data to be donated to jax,
+            # Note that we mark the state to be donated to jax,
             # so that jax will reuse the memory buffer for outputs.
             # This will reduce the memory usage of the training function by
             # half.
-            @partial(jax.jit, donate_argnames="state")
-            def compiled_train_step(state, data):
-                return train_step(state, data)
-
-            self.train_function = compiled_train_step
-
+            train_step = jax.jit(self.train_step, donate_argnums=0)
         else:
-            self.train_function = train_step
+            train_step = self.train_step
+
+        step_function = self._make_function(train_step)
+
+        self.train_function = step_function
 
     def make_test_function(self, force=False):
         if self.test_function is not None and not force:
             return
-
-        def one_test_step(state, data):
-            data = data[0]
-            return self.test_step(state, data)
-
-        def multi_test_steps(state, data):
-            for single_step_data in data:
-                logs, state = one_test_step(state, [single_step_data])
-            return logs, state
-
-        if self.steps_per_execution > 1:
-            test_step = multi_test_steps
-        else:
-            test_step = one_test_step
-
         if not self.run_eagerly and self.jit_compile:
-            # Note that we mark the state and data to be donated to jax,
+            # Note that we mark the state to be donated to jax,
             # so that jax will reuse the memory buffer for outputs.
             # This will reduce the memory usage of the training function by
             # half.
-            @partial(jax.jit, donate_argnames="state")
-            def compiled_test_step(state, data):
-                return test_step(state, data)
-
-            self.test_function = compiled_test_step
-
+            test_step = jax.jit(self.test_step, donate_argnums=0)
         else:
-            self.test_function = test_step
+            test_step = self.test_step
+
+        step_function = self._make_function(test_step)
+
+        self.test_function = step_function
 
     def make_predict_function(self, force=False):
         if self.predict_function is not None and not force:
             return self.predict_function
 
-        def one_predict_step(state, data):
-            data = data[0]
-            return self.predict_step(state, data)
-
-        def multi_predict_steps(state, data):
-            outputs, trainable_variables = one_predict_step(state, data[:1])
-
-            for single_step_data in data[1:]:
-                step_outputs, trainable_variables = one_predict_step(
-                    state,
-                    [single_step_data],
-                )
-                outputs = tree.map_structure(
-                    lambda t1, t2: jax.numpy.concatenate([t1, t2]),
-                    outputs,
-                    step_outputs,
-                )
-            return outputs, trainable_variables
-
-        if self.steps_per_execution > 1:
-            predict_step = multi_predict_steps
-        else:
-            predict_step = one_predict_step
+        def predict_step(state, data):
+            outputs, non_trainable_variables = self.predict_step(state, data)
+            return outputs, (state[0], non_trainable_variables)
 
         if not self.run_eagerly and self.jit_compile:
+            predict_step = jax.jit(predict_step, donate_argnums=0)
 
-            @jax.jit
-            def compiled_predict_step(state, data):
-                return predict_step(state, data)
+        _step_function = self._make_function(
+            predict_step, concatenate_outputs=True
+        )
 
-            self.predict_function = compiled_predict_step
+        def step_function(state, iterator):
+            outputs, state = _step_function(state, iterator)
+            return outputs, state
 
-        else:
-            self.predict_function = predict_step
+        self.predict_function = step_function
 
     @traceback_utils.filter_traceback
     def fit(
@@ -354,10 +347,9 @@ class JAXTrainer(base_trainer.Trainer):
             # Create the validation data using the training data. Only supported
             # for TF/numpy/jax arrays.
             (
-                x,
-                y,
-                sample_weight,
-            ), validation_data = array_slicing.train_validation_split(
+                (x, y, sample_weight),
+                validation_data,
+            ) = array_slicing.train_validation_split(
                 (x, y, sample_weight), validation_split=validation_split
             )
 
@@ -381,6 +373,7 @@ class JAXTrainer(base_trainer.Trainer):
         )
 
         self._symbolic_build(iterator=epoch_iterator)
+        epoch_iterator.reset()
 
         # Container that configures and calls callbacks.
         if not isinstance(callbacks, callbacks_module.CallbackList):
@@ -397,6 +390,7 @@ class JAXTrainer(base_trainer.Trainer):
 
         self.make_train_function()
         self.stop_training = False
+        training_logs = {}
         callbacks.on_train_begin()
         initial_epoch = self._initial_epoch or initial_epoch
         for epoch in range(initial_epoch, epochs):
@@ -404,44 +398,46 @@ class JAXTrainer(base_trainer.Trainer):
             callbacks.on_epoch_begin(epoch)
 
             self._jax_state_synced = True
-            for step, data in epoch_iterator.enumerate_epoch():
-                # Callbacks
-                callbacks.on_train_batch_begin(step)
+            with epoch_iterator.catch_stop_iteration():
+                for step, iterator in epoch_iterator:
+                    # Callbacks
+                    callbacks.on_train_batch_begin(step)
 
-                # Train step
-                if self._jax_state_synced:
-                    # The state may have been synced by a callback.
-                    state = self._get_jax_state(
-                        trainable_variables=True,
-                        non_trainable_variables=True,
-                        optimizer_variables=True,
-                        metrics_variables=True,
-                        purge_model_variables=True,
-                    )
-                    self._jax_state_synced = False
+                    # Train step
+                    if self._jax_state_synced:
+                        # The state may have been synced by a callback.
+                        state = self._get_jax_state(
+                            trainable_variables=True,
+                            non_trainable_variables=True,
+                            optimizer_variables=True,
+                            metrics_variables=True,
+                            purge_model_variables=True,
+                        )
+                        self._jax_state_synced = False
 
-                logs, state = self.train_function(state, data)
-                (
-                    trainable_variables,
-                    non_trainable_variables,
-                    optimizer_variables,
-                    metrics_variables,
-                ) = state
+                    logs, state = self.train_function(state, iterator)
+                    (
+                        trainable_variables,
+                        non_trainable_variables,
+                        optimizer_variables,
+                        metrics_variables,
+                    ) = state
 
-                # Setting _jax_state enables callbacks to force a state sync
-                # if they need to.
-                self._jax_state = {
-                    "trainable_variables": trainable_variables,
-                    "non_trainable_variables": non_trainable_variables,
-                    "optimizer_variables": optimizer_variables,
-                    "metrics_variables": metrics_variables,
-                }
+                    # Setting _jax_state enables callbacks to force a state sync
+                    # if they need to.
+                    self._jax_state = {
+                        "trainable_variables": trainable_variables,
+                        "non_trainable_variables": non_trainable_variables,
+                        "optimizer_variables": optimizer_variables,
+                        "metrics_variables": metrics_variables,
+                    }
+                    # Dispatch callbacks. This takes care of async dispatch.
+                    callbacks.on_train_batch_end(step, logs)
 
-                # Callbacks
-                logs = self._pythonify_logs(logs)
-                callbacks.on_train_batch_end(step, logs)
-                if self.stop_training:
-                    break
+                    if self.stop_training:
+                        # Stop training if a callback has set
+                        # this flag in on_(train_)batch_end.
+                        break
 
             # Reattach state to the model (if not already done by a callback).
             # NOTE: doing this after each step would be a big performance
@@ -501,6 +497,7 @@ class JAXTrainer(base_trainer.Trainer):
             del self._eval_epoch_iterator
         callbacks.on_train_end(logs=training_logs)
         self._jax_state = None
+        self._clear_jax_state_sharding()
         return self.history
 
     @traceback_utils.filter_traceback
@@ -537,12 +534,12 @@ class JAXTrainer(base_trainer.Trainer):
             )
 
         self._symbolic_build(iterator=epoch_iterator)
+        epoch_iterator.reset()
 
         # Container that configures and calls callbacks.
         if not isinstance(callbacks, callbacks_module.CallbackList):
             callbacks = callbacks_module.CallbackList(
                 callbacks,
-                add_history=True,
                 add_progbar=verbose != 0,
                 verbose=verbose,
                 epochs=1,
@@ -554,43 +551,46 @@ class JAXTrainer(base_trainer.Trainer):
         self.make_test_function()
         self.stop_evaluating = False
         callbacks.on_test_begin()
-        logs = None
+        logs = {}
         self.reset_metrics()
 
         self._jax_state_synced = True
-        for step, data in epoch_iterator.enumerate_epoch():
-            callbacks.on_test_batch_begin(step)
+        with epoch_iterator.catch_stop_iteration():
+            for step, iterator in epoch_iterator:
+                callbacks.on_test_batch_begin(step)
 
-            if self._jax_state_synced:
-                # The state may have been synced by a callback.
-                state = self._get_jax_state(
-                    trainable_variables=True,
-                    non_trainable_variables=True,
-                    metrics_variables=True,
-                    purge_model_variables=True,
-                )
-                self._jax_state_synced = False
+                if self._jax_state_synced:
+                    # The state may have been synced by a callback.
+                    state = self._get_jax_state(
+                        trainable_variables=True,
+                        non_trainable_variables=True,
+                        metrics_variables=True,
+                        purge_model_variables=True,
+                    )
+                    self._jax_state_synced = False
 
-            logs, state = self.test_function(state, data)
-            (
-                trainable_variables,
-                non_trainable_variables,
-                metrics_variables,
-            ) = state
+                logs, state = self.test_function(state, iterator)
+                (
+                    trainable_variables,
+                    non_trainable_variables,
+                    metrics_variables,
+                ) = state
 
-            # Setting _jax_state enables callbacks to force a state sync
-            # if they need to.
-            self._jax_state = {
-                # I wouldn't recommend modifying non-trainable model state
-                # during evaluate(), but it's allowed.
-                "trainable_variables": trainable_variables,
-                "non_trainable_variables": non_trainable_variables,
-                "metrics_variables": metrics_variables,
-            }
-            logs = self._pythonify_logs(logs)
-            callbacks.on_test_batch_end(step, logs)
-            if self.stop_evaluating:
-                break
+                # Setting _jax_state enables callbacks to force a state sync
+                # if they need to.
+                self._jax_state = {
+                    # I wouldn't recommend modifying non-trainable model state
+                    # during evaluate(), but it's allowed.
+                    "trainable_variables": trainable_variables,
+                    "non_trainable_variables": non_trainable_variables,
+                    "metrics_variables": metrics_variables,
+                }
+
+                # Dispatch callbacks. This takes care of async dispatch.
+                callbacks.on_test_batch_end(step, logs)
+
+                if self.stop_evaluating:
+                    break
 
         # Reattach state back to model (if not already done by a callback).
         self.jax_state_sync()
@@ -602,6 +602,9 @@ class JAXTrainer(base_trainer.Trainer):
             logs = self._get_metrics_result_or_logs(logs)
         callbacks.on_test_end(logs)
         self._jax_state = None
+        if not use_cached_eval_dataset:
+            # Only clear sharding if evaluate is not called from `fit`.
+            self._clear_jax_state_sharding()
         if return_dict:
             return logs
         return self._flatten_metrics_in_order(logs)
@@ -621,18 +624,19 @@ class JAXTrainer(base_trainer.Trainer):
 
         if not all(layer.built for layer in self._flatten_layers()):
             # Build the model on one batch of data.
-            for _, data in epoch_iterator.enumerate_epoch():
+            for _, iterator in epoch_iterator:
                 # Build model
-                x, _, _ = data_adapter_utils.unpack_x_y_sample_weight(data[0])
+                x, _, _ = data_adapter_utils.unpack_x_y_sample_weight(
+                    next(iterator)
+                )
                 with backend.StatelessScope():
                     self(x)
                 break
-
+            epoch_iterator.reset()
         # Container that configures and calls callbacks.
         if not isinstance(callbacks, callbacks_module.CallbackList):
             callbacks = callbacks_module.CallbackList(
                 callbacks,
-                add_history=True,
                 add_progbar=verbose != 0,
                 verbose=verbose,
                 epochs=1,
@@ -663,34 +667,40 @@ class JAXTrainer(base_trainer.Trainer):
         self._jax_state_synced = True
         outputs = None
         non_trainable_variables = None
-        for step, x in epoch_iterator.enumerate_epoch():
-            callbacks.on_predict_batch_begin(step)
-            if self._jax_state_synced:
-                # The state may have been synced by a callback.
-                state = self._get_jax_state(
-                    trainable_variables=True,
-                    non_trainable_variables=True,
-                )
-                self._purge_model_variables(non_trainable_variables=True)
-                self._jax_state_synced = False
-            else:
-                state = (state[0], non_trainable_variables)
-            batch_outputs, non_trainable_variables = self.predict_function(
-                state, x
-            )
-            outputs = append_to_outputs(batch_outputs, outputs)
-            callbacks.on_predict_batch_end(step, {"outputs": batch_outputs})
-            if self.stop_predicting:
-                break
+        with epoch_iterator.catch_stop_iteration():
+            for step, iterator in epoch_iterator:
+                callbacks.on_predict_batch_begin(step)
+                if self._jax_state_synced:
+                    # The state may have been synced by a callback.
+                    state = self._get_jax_state(
+                        trainable_variables=True,
+                        non_trainable_variables=True,
+                        purge_model_variables=True,
+                    )
+                    self._jax_state_synced = False
+                batch_outputs, state = self.predict_function(state, iterator)
+                (
+                    trainable_variables,
+                    non_trainable_variables,
+                ) = state
+                self._jax_state = {
+                    "trainable_variables": trainable_variables,
+                    # I wouldn't recommend modifying non-trainable model state
+                    # during predict(), but it's allowed.
+                    "non_trainable_variables": non_trainable_variables,
+                }
+                outputs = append_to_outputs(batch_outputs, outputs)
+
+                # Dispatch callbacks. This takes care of async dispatch.
+                callbacks.on_predict_batch_end(step, {"outputs": batch_outputs})
+
+                if self.stop_predicting:
+                    break
 
-        self._jax_state = {
-            # I wouldn't recommend modifying non-trainable model state
-            # during predict(), but it's allowed.
-            "non_trainable_variables": non_trainable_variables,
-        }
         self.jax_state_sync()
         callbacks.on_predict_end()
         self._jax_state = None
+        self._clear_jax_state_sharding()
         return tree.map_structure_up_to(batch_outputs, np.concatenate, outputs)
 
     def train_on_batch(
@@ -713,11 +723,12 @@ class JAXTrainer(base_trainer.Trainer):
             sample_weight = data_adapter_utils.class_weight_to_sample_weights(
                 y, class_weight
             )
-        data = (x, y, sample_weight)
-        data = _distribute_data(data)
+
+        def data():
+            yield _distribute_data((x, y, sample_weight))
 
         # Maybe build model
-        self._symbolic_build(data_batch=data)
+        self._symbolic_build(data_batch=next(data()))
         self._record_training_state_sharding_spec()
         self.make_train_function()
 
@@ -730,7 +741,7 @@ class JAXTrainer(base_trainer.Trainer):
             purge_model_variables=False,
         )
         self._jax_state_synced = False
-        logs, state = self.train_function(state, [data])
+        logs, state = self.train_function(state, data())
 
         # State sync
         (
@@ -762,10 +773,11 @@ class JAXTrainer(base_trainer.Trainer):
     ):
         self._assert_compile_called("test_on_batch")
 
-        data = (x, y, sample_weight)
-        data = _distribute_data(data)
+        def data():
+            yield _distribute_data((x, y, sample_weight))
+
         # Maybe build model
-        self._symbolic_build(data_batch=data)
+        self._symbolic_build(data_batch=next(data()))
         self._record_training_state_sharding_spec()
         self.make_test_function()
 
@@ -777,7 +789,7 @@ class JAXTrainer(base_trainer.Trainer):
             purge_model_variables=False,
         )
         self._jax_state_synced = False
-        logs, state = self.test_function(state, [data])
+        logs, state = self.test_function(state, data())
 
         # State sync
         trainable_variables, non_trainable_variables, metrics_variables = state
@@ -809,10 +821,14 @@ class JAXTrainer(base_trainer.Trainer):
             purge_model_variables=False,
         )
         self._jax_state_synced = False
-        batch_outputs, non_trainable_variables = self.predict_function(
-            state, [(x,)]
-        )
+
+        def data():
+            yield (x,)
+
+        batch_outputs, state = self.predict_function(state, data())
+        trainable_variables, non_trainable_variables = state
         self._jax_state = {
+            "trainable_variables": trainable_variables,
             "non_trainable_variables": non_trainable_variables,
         }
         self.jax_state_sync()
@@ -862,6 +878,12 @@ class JAXTrainer(base_trainer.Trainer):
             v.value.sharding for v in self.metrics_variables
         ]
 
+    def _clear_jax_state_sharding(self):
+        self._trainable_variable_shardings = None
+        self._non_trainable_variable_shardings = None
+        self._optimizer_variable_shardings = None
+        self._metrics_variable_shardings = None
+
     def _enforce_jax_state_sharding(
         self,
         trainable_variables=None,
@@ -919,9 +941,9 @@ class JAXTrainer(base_trainer.Trainer):
     ):
         """Remove all the model variable for memory saving.
 
-        During JAX training, since the training function are stateless, we have
+        During JAX training, since the training function is stateless, we have
         to pass in and get the model weights over and over, during which the
-        copy of the weights that attached to the KerasVariable are still and
+        copy of the weights that attached to the Variable are still and
         occupying extra memory. We remove those variable to save memory (for
         better memory utilization) at the beginning of the epoch, and reattach
         the value back to variables at the end of the epoch, via
@@ -967,25 +989,50 @@ class JAXTrainer(base_trainer.Trainer):
         return tuple(state)
 
 
-def _distribute_data(data):
+def _distribute_data(data, layouts=None):
     distribution = distribution_lib.distribution()
+
     if distribution is not None:
+        if layouts is None:
+            layouts = tree.map_structure(
+                lambda d: distribution.get_data_layout(d.shape),
+                data,
+            )
+        jax_dist_data_input = partial(
+            jax_distribution_lib.distribute_data_input,
+            batch_dim_name=distribution.batch_dim_name,
+        )
+        return tree.map_structure(jax_dist_data_input, data, layouts)
 
-        def distribute_single_value(d):
-            layout = distribution.get_data_layout(d.shape)
-            return jax_distribution_lib.distribute_data_input(d, layout)
-
-        return tree.map_structure(distribute_single_value, data)
-    else:
-        return tree.map_structure(jax.device_put, data)
+    return tree.map_structure(jax.device_put, data)
 
 
 class JAXEpochIterator(EpochIterator):
+    def __next__(self):
+        return next(self._epoch_iterator)
+
     def _get_iterator(self):
+        distribution = distribution_lib.distribution()
+        if distribution is not None:
+            return self._get_distributed_iterator(distribution)
+
         return self._prefetch_numpy_iterator(
             self.data_adapter.get_jax_iterator()
         )
 
+    def _get_distributed_iterator(self, distribution):
+        """Lazily compute layouts to reduce host to device transfer latency."""
+        layouts = None
+        for data in self.data_adapter.get_jax_iterator():
+            if layouts is None:
+                layouts = tree.map_structure(
+                    lambda d: jax_distribution_lib._to_jax_layout(
+                        distribution.get_data_layout(d.shape)
+                    ),
+                    data,
+                )
+            yield _distribute_data(data, layouts)
+
     def _prefetch_numpy_iterator(self, numpy_iterator):
         """Shard and prefetch batches on device.
 
