@@ -1,9 +1,7 @@
-"""Unified high level distribution APIs across backends.
+"""Unified high-level distribution APIs across backends.
 
-!!!DO NOT USE!!! Currently under development and APIs are not final.
-
-Currently only the JAX backend has been implemented. The TensorFlow backend
-will be implemented in the future (via tf.dtensor API).
+Currently only the JAX backend is supported. The TensorFlow backend
+will be supported in the future (via tf.dtensor API).
 """
 
 import collections
@@ -98,7 +96,7 @@ def initialize(job_addresses=None, num_processes=None, process_id=None):
         ```python
         os.environ[
             "KERAS_DISTRIBUTION_JOB_ADDRESSES"] = "10.0.0.1:1234,10.0.0.2:2345"
-        os.environ["KERAS_DISTRIBUTION_NUM_PROCESSES"] = "2
+        os.environ["KERAS_DISTRIBUTION_NUM_PROCESSES"] = "2"
         os.environ["KERAS_DISTRIBUTION_PROCESS_ID"] = "0"
         keras.distribute.initialize()
         ```
@@ -107,7 +105,7 @@ def initialize(job_addresses=None, num_processes=None, process_id=None):
         ```python
         os.environ[
             "KERAS_DISTRIBUTION_JOB_ADDRESSES"] = "10.0.0.1:1234,10.0.0.2:2345"
-        os.environ["KERAS_DISTRIBUTION_NUM_PROCESSES"] = "2
+        os.environ["KERAS_DISTRIBUTION_NUM_PROCESSES"] = "2"
         os.environ["KERAS_DISTRIBUTION_PROCESS_ID"] = "1"
         keras.distribute.initialize()
         ```
@@ -289,8 +287,9 @@ class Distribution:
         device_mesh: A `DeviceMesh` instance.
     """
 
-    def __init__(self, device_mesh):
+    def __init__(self, device_mesh, batch_dim_name=None):
         self._device_mesh = device_mesh
+        self._batch_dim_name = batch_dim_name
 
     def get_data_layout(self, data_shape):
         """Retrieve the `TensorLayout` for the input data.
@@ -308,7 +307,7 @@ class Distribution:
         """Retrieve the `TensorLayout` for the variable.
 
         Args:
-            variable: A `KerasVariable` instance.
+            variable: A `Variable` instance.
 
         return:
             The `TensorLayout` for the variable, which can be used by
@@ -343,6 +342,10 @@ class Distribution:
     def device_mesh(self):
         return self._device_mesh
 
+    @property
+    def batch_dim_name(self):
+        return self._batch_dim_name
+
     def distribute_dataset(self, dataset):
         """Create a distributed dataset instance from the original user dataset.
 
@@ -385,9 +388,11 @@ class DataParallel(Distribution):
     Args:
         device_mesh: Optional `DeviceMesh` instance.
         devices: Optional list of devices.
+        auto_shard_dataset: Automatically shard the dataset amongst processes.
+            Defaults to true.
     """
 
-    def __init__(self, device_mesh=None, devices=None):
+    def __init__(self, device_mesh=None, devices=None, auto_shard_dataset=True):
         if device_mesh:
             self._initialize_with_device_mesh(device_mesh)
         elif devices:
@@ -395,11 +400,11 @@ class DataParallel(Distribution):
         else:
             self._initialize_mesh_from_list_devices()
 
-        self._batch_dim_name = self.device_mesh.axis_names[0]
         # Those following attributes might get convert to public methods.
         self._num_process = distribution_lib.num_processes()
         self._process_id = distribution_lib.process_id()
         self._is_multi_process = self._num_process > 1
+        self._auto_shard_dataset = auto_shard_dataset
 
     def _initialize_with_device_mesh(self, device_mesh):
         if not isinstance(device_mesh, DeviceMesh):
@@ -407,7 +412,7 @@ class DataParallel(Distribution):
                 "Expect `mesh` to be an instance of `DeviceMesh`. "
                 f"Received: mesh={device_mesh} (of type {type(device_mesh)})"
             )
-        super().__init__(device_mesh)
+        super().__init__(device_mesh, device_mesh.axis_names[0])
         if self.device_mesh.devices.ndim != 1:
             warnings.warn(
                 "Expect the input mesh to be 1D, but received "
@@ -423,7 +428,7 @@ class DataParallel(Distribution):
             axis_names=[DEFAULT_BATCH_DIM_NAME],
             devices=devices,
         )
-        super().__init__(device_mesh)
+        super().__init__(device_mesh, DEFAULT_BATCH_DIM_NAME)
 
     def _initialize_mesh_from_list_devices(self):
         devices = np.array(list_devices())
@@ -432,11 +437,11 @@ class DataParallel(Distribution):
             axis_names=[DEFAULT_BATCH_DIM_NAME],
             devices=devices,
         )
-        super().__init__(device_mesh)
+        super().__init__(device_mesh, DEFAULT_BATCH_DIM_NAME)
 
     def get_data_layout(self, data_shape):
         data_shard_spec = [None] * len(data_shape)
-        data_shard_spec[0] = self._batch_dim_name  # Shard on the first dim
+        data_shard_spec[0] = self.batch_dim_name  # Shard on the first dim
         return TensorLayout(data_shard_spec, self.device_mesh)
 
     def get_variable_layout(self, variable):
@@ -459,7 +464,7 @@ class DataParallel(Distribution):
                 "Only `tf.data.Dataset` is supported for "
                 f"sharding, got {type(dataset)}"
             )
-        if not self._is_multi_process:
+        if not self._is_multi_process or not self._auto_shard_dataset:
             return dataset
 
         batch_size = tf_data_distribute.compute_batch_size(dataset)
@@ -589,7 +594,7 @@ class ModelParallel(Distribution):
 
     def get_data_layout(self, data_shape):
         data_shard_spec = [None] * len(data_shape)
-        data_shard_spec[0] = self._batch_dim_name  # Shard on the first dim
+        data_shard_spec[0] = self.batch_dim_name  # Shard on the first dim
         return TensorLayout(data_shard_spec, self.device_mesh)
 
     def get_variable_layout(self, variable):
@@ -625,63 +630,54 @@ class ModelParallel(Distribution):
                 "the input dataset, e.g via `dataset.batch(batch_size)`"
             )
 
-        # We need to compute the the per-worker batch size.
-        # Imagine we have 4 workers, and each worker has 2 devices. The device
-        # ID will be w{worker_id}_{device_id} from global perspective, eg
-        # 'w0_0', 'w0_1', 'w1_0', etc.
-        # For a mesh with global shape {'batch': 4, 'model': 2}, we would like
-        # the mesh to be like below:
-        #         batch ->
-        #         --------------------------
-        # model   | w0_0, w0_1, w1_0, w1_1 |
-        #   v     | w0_2, w0_3, w1_2, w1_3 |
-        #         --------------------------
-        # In this case, the batch dim will be split across 2 workers.
-        #
-        # In the case that global batch dim is smaller than the worker size, eg
-        # {'batch': 1, 'model': 8}, then the mesh will be like below:
-        #         batch ->
-        #         --------
-        # model   | w0_0 |
-        #         | w0_1 |
-        #         | w1_0 |
-        #         | w1_1 |
-        #         | w2_0 |
-        #         | w2_1 |
-        #         | w3_0 |
-        #         | w3_1 |
-        #         --------
-        # And the data batch will be fully replicated for each of the worker.
+        # We need to compute the per-process/worker/host batch size.
+        # This will depend on how many model replicas we have on each process.
+        # Note that this might be smaller than one if model replicas are sharded
+        # across multiple processes.
         mesh_batch_dim_index = self.device_mesh.axis_names.index(
-            self._batch_dim_name
+            self.batch_dim_name
         )
-        mesh_batch_dim_size = self.device_mesh.shape[mesh_batch_dim_index]
-        # TODO(scottzhu): local device count can be a field of the mesh.
-        local_device_count = (
-            np.prod(self.device_mesh.shape) // self._num_process
-        )
-        if mesh_batch_dim_size < local_device_count:
-            # No sharding is needed in this case. The worker will have the
+        num_model_replicas = self.device_mesh.shape[mesh_batch_dim_index]
+        if num_model_replicas == 1:
+            # No sharding is needed in this case. Each process will have the
             # global batch size, and data from the iterator will need to be
-            # replicated for model dimension.
+            # replicated across all processes.
             return dataset.prefetch(tf.data.AUTOTUNE)
-        else:
-            if mesh_batch_dim_size % local_device_count != 0:
+        num_model_replicas_per_process = num_model_replicas / self._num_process
+        if num_model_replicas_per_process >= 1:
+            # Each process will have one or more full model replicas. Data will
+            # be sharded across all processes without replication.
+            if global_batch_size % self._num_process != 0:
                 raise ValueError(
-                    "The Batch dimension of the mesh is not compatible "
-                    "with the local worker device count. Mesh batch "
-                    f"dim = {mesh_batch_dim_size} and local device "
-                    f"count = {local_device_count}"
+                    "Global batch size must be divisible by the number of "
+                    f"processes. `global_batch_size`={global_batch_size} and "
+                    f"`num_process`={self._num_process}"
                 )
-            num_shards = mesh_batch_dim_size // local_device_count
-            per_worker_batch_size = global_batch_size // num_shards
-
-            distributed_dataset = dataset.rebatch(per_worker_batch_size)
-            distributed_dataset = tf_data_distribute._AutoShardDataset(
-                distributed_dataset,
-                num_workers=num_shards,
-                index=self._process_id % num_shards,
-                num_replicas=num_shards,
+            per_process_batch_size = global_batch_size // self._num_process
+            distributed_dataset = dataset.rebatch(per_process_batch_size)
+            distributed_dataset = distributed_dataset.shard(
+                num_shards=self._num_process,
+                index=self._process_id,
+            )
+            return distributed_dataset.prefetch(tf.data.AUTOTUNE)
+        else:
+            # Model replicas are sharded across multiple processes. Data will be
+            # sharded across model replicas, and replicated across processes
+            # within the same model replica.
+            if global_batch_size % num_model_replicas != 0:
+                raise ValueError(
+                    "Global batch size must be divisible by the number of "
+                    f"replicas. `global_batch_size`={global_batch_size} and "
+                    f"`num_model_replicas`={num_model_replicas}"
+                )
+            per_process_batch_size = global_batch_size // num_model_replicas
+            distributed_dataset = dataset.rebatch(per_process_batch_size)
+            processes_per_replica = self._num_process // num_model_replicas
+            # TODO: Figure out what the convention is for data sharding id.
+            data_shard_id = self._process_id % processes_per_replica
+            distributed_dataset = distributed_dataset.shard(
+                num_shards=num_model_replicas,
+                index=data_shard_id,
             )
             return distributed_dataset.prefetch(tf.data.AUTOTUNE)
 
@@ -700,7 +696,7 @@ class LayoutMap(collections.abc.MutableMapping):
     `TensorLayout` instance.
 
     In the normal case, the key to query is usually the `variable.path`, which
-    is the idenifier of the variable.
+    is the identifier of the variable.
 
     As shortcut, tuple or list of axis names are also allowed when inserting
     as value, and will be converted to `TensorLayout`.
