@@ -220,7 +220,7 @@ class BaseDepthwiseConv(Layer):
                     1,
                 ) * self.rank
             bias = ops.reshape(self.bias, bias_shape)
-            outputs += bias
+            outputs = ops.add(outputs, bias)
 
         if self.activation is not None:
             return self.activation(outputs)
