@@ -224,8 +224,7 @@ class EinsumDense(Layer):
             )
         if self.lora_enabled:
             raise ValueError(
-                "lora is already enabled. "
-                "This can only be done once per layer."
+                "lora is already enabled. This can only be done once per layer."
             )
         self._tracker.unlock()
         self.lora_kernel_a = self.add_weight(
@@ -360,13 +359,6 @@ class EinsumDense(Layer):
 
     # Quantization-related (int8 and float8) methods
 
-    def _quantization_mode_error(self, mode):
-        return NotImplementedError(
-            "Invalid quantization mode. Expected one of "
-            f"{dtype_policies.QUANTIZATION_MODES}. "
-            f"Received: quantization_mode={mode}"
-        )
-
     def quantized_build(self, input_shape, mode):
         if mode == "int8":
             shape_data = _analyze_einsum_string(
@@ -477,15 +469,7 @@ class EinsumDense(Layer):
         self.outputs_grad_amax_history.overwrite_with_gradient = True
         self._is_quantized = True
 
-    def quantized_call(self, inputs, training=None):
-        if self.quantization_mode == "int8":
-            return self._int8_call(inputs)
-        elif self.quantization_mode == "float8":
-            return self._float8_call(inputs, training=training)
-        else:
-            raise self._quantization_mode_error(self.quantization_mode)
-
-    def _int8_call(self, inputs):
+    def _int8_call(self, inputs, training=None):
         @ops.custom_gradient
         def einsum_with_inputs_gradient(inputs, kernel, kernel_scale):
             def grad_fn(*args, upstream=None):
@@ -640,18 +624,11 @@ class EinsumDense(Layer):
             x = self.activation(x)
         return x
 
-    def quantize(self, mode):
-        import gc
-
+    def quantize(self, mode, type_check=True):
         # Prevent quantization of the subclasses
-        if type(self) is not EinsumDense:
-            raise NotImplementedError(
-                f"Layer {self.__class__.__name__} does not have a `quantize()` "
-                "method implemented."
-            )
-        self._check_quantize_args(mode, self.compute_dtype)
+        if type_check and (type(self) is not EinsumDense):
+            raise self._not_implemented_error(self.quantize)
 
-        self._tracker.unlock()
         if mode == "int8":
             (
                 self._input_reduced_axes,
@@ -667,7 +644,7 @@ class EinsumDense(Layer):
             ) = _analyze_quantization_info(self.equation, self.input_spec.ndim)
             # Quantize `self._kernel` to int8 and compute corresponding scale
             kernel_value, kernel_scale = quantizers.abs_max_quantize(
-                self._kernel, axis=self._kernel_reduced_axes
+                self._kernel, axis=self._kernel_reduced_axes, to_numpy=True
             )
             kernel_scale = ops.transpose(
                 kernel_scale, self._kernel_transpose_axes
@@ -680,30 +657,21 @@ class EinsumDense(Layer):
                 kernel_scale = ops.squeeze(
                     kernel_scale, axis=self._kernel_squeeze_axes
                 )
-            self._untrack_variable(self._kernel)
-            kernel_shape = self._kernel.shape
+            kernel_shape = tuple(self._kernel.shape)
             del self._kernel
             # Utilize a lambda expression as an initializer to prevent adding a
             # large constant to the computation graph.
-            self._int8_build(
-                kernel_shape,
-                lambda shape, dtype: kernel_value,
-                lambda shape, dtype: kernel_scale,
-            )
+            self._int8_build(kernel_shape, kernel_value, kernel_scale)
         elif mode == "float8":
             self._float8_build()
         else:
             raise self._quantization_mode_error(mode)
-        self._tracker.lock()
 
         # Set new dtype policy
         if self.dtype_policy.quantization_mode is None:
             policy = dtype_policies.get(f"{mode}_from_{self.dtype_policy.name}")
             self.dtype_policy = policy
 
-        # Release memory manually because sometimes the backend doesn't
-        gc.collect()
-
     def _get_kernel_with_merged_lora(self):
         if self.dtype_policy.quantization_mode is not None:
             kernel_value = self._kernel
@@ -711,13 +679,31 @@ class EinsumDense(Layer):
             if self.lora_enabled:
                 # Dequantize & quantize to merge lora weights into int8 kernel
                 # Note that this is a lossy compression
+                if self._kernel_squeeze_axes:
+                    kernel_scale = ops.expand_dims(
+                        kernel_scale, axis=self._kernel_squeeze_axes
+                    )
+                if self._kernel_expand_axes:
+                    kernel_scale = ops.squeeze(
+                        kernel_scale, axis=self._kernel_expand_axes
+                    )
+                if self._kernel_transpose_axes:
+
+                    def _argsort(seq):
+                        # Ref: https://stackoverflow.com/a/3382369
+                        return sorted(range(len(seq)), key=seq.__getitem__)
+
+                    reverse_transpose = _argsort(self._kernel_transpose_axes)
+                    kernel_scale = ops.transpose(
+                        kernel_scale, axes=reverse_transpose
+                    )
                 kernel_value = ops.divide(kernel_value, kernel_scale)
                 kernel_value = ops.add(
                     kernel_value,
                     ops.matmul(self.lora_kernel_a, self.lora_kernel_b),
                 )
                 kernel_value, kernel_scale = quantizers.abs_max_quantize(
-                    kernel_value, axis=self._kernel_reduced_axes
+                    kernel_value, axis=self._kernel_reduced_axes, to_numpy=True
                 )
                 kernel_scale = ops.transpose(
                     kernel_scale, self._kernel_transpose_axes
@@ -886,7 +872,6 @@ def _analyze_split_string(
 
 
 def _analyze_quantization_info(equation, input_shape):
-
     def get_specs(equation, input_shape):
         possible_labels = string.ascii_letters
         dot_replaced_string = re.sub(r"\.\.\.", "0", equation)
