@@ -17,7 +17,9 @@ And some more magic:
 """
 
 import collections
+import functools
 import inspect
+import math
 import warnings
 from functools import wraps
 
@@ -31,7 +33,10 @@ from keras.src import utils
 from keras.src.api_export import keras_export
 from keras.src.backend import KerasTensor
 from keras.src.backend.common import global_state
+from keras.src.backend.common import remat
 from keras.src.backend.common.name_scope import current_path
+from keras.src.backend.common.remat import get_current_remat_mode
+from keras.src.backend.common.symbolic_scope import in_symbolic_scope
 from keras.src.distribution import distribution_lib
 from keras.src.dtype_policies import DTypePolicyMap
 from keras.src.layers import input_spec
@@ -51,6 +56,8 @@ elif backend.backend() == "torch":
     from keras.src.backend.torch.layer import TorchLayer as BackendLayer
 elif backend.backend() == "numpy":
     from keras.src.backend.numpy.layer import NumpyLayer as BackendLayer
+elif backend.backend() == "openvino":
+    from keras.src.backend.openvino.layer import OpenvinoLayer as BackendLayer
 else:
     raise RuntimeError(
         f"Backend '{backend.backend()}' must implement a layer mixin class."
@@ -82,12 +89,10 @@ class Layer(BackendLayer, Operation, KerasSaveable):
         trainable: Boolean, whether the layer's variables should be trainable.
         name: String name of the layer.
         dtype: The dtype of the layer's computations and weights. Can also be a
-            `keras.DTypePolicy`,
-            which allows the computation and
-            weight dtype to differ. Defaults to `None`. `None` means to use
-            `keras.config.dtype_policy()`,
-            which is a `float32` policy unless set to different value
-            (via `keras.config.set_dtype_policy()`).
+            `keras.DTypePolicy`, which allows the computation and weight dtype
+            to differ. Defaults to `None`. `None` means to use
+            `keras.config.dtype_policy()`, which is a `float32` policy unless
+            set to different value (via `keras.config.set_dtype_policy()`).
 
     Attributes:
         name: The name of the layer (string).
@@ -212,10 +217,10 @@ class Layer(BackendLayer, Operation, KerasSaveable):
     """
 
     def __new__(cls, *args, **kwargs):
-        # Wrap the user-provided build method in the build_decorator
-        # to add name scope support and serialization support.
         obj = super().__new__(cls, *args, **kwargs)
 
+        # Wrap the user-provided `build` method in the `build_wrapper`
+        # to add name scope support and serialization support.
         original_build_method = obj.build
 
         @wraps(original_build_method)
@@ -232,6 +237,24 @@ class Layer(BackendLayer, Operation, KerasSaveable):
             obj._lock_state()
 
         obj.build = build_wrapper
+
+        # Wrap the user-provided `quantize` method in the `quantize_wrapper`
+        # to add tracker support.
+        original_quantize_method = obj.quantize
+
+        @wraps(original_quantize_method)
+        def quantize_wrapper(mode, **kwargs):
+            obj._check_quantize_args(mode, obj.compute_dtype)
+            obj._tracker.unlock()
+            try:
+                original_quantize_method(mode, **kwargs)
+            except Exception:
+                raise
+            finally:
+                obj._tracker.lock()
+
+        obj.quantize = quantize_wrapper
+
         return obj
 
     def __init__(
@@ -296,6 +319,7 @@ class Layer(BackendLayer, Operation, KerasSaveable):
         self._build_shapes_dict = None
         # Parent path
         self._parent_path = None
+        self._remat_mode = get_current_remat_mode()
         self._initialize_tracker()
 
     @tracking.no_automatic_dependency_tracking
@@ -472,7 +496,7 @@ class Layer(BackendLayer, Operation, KerasSaveable):
         autocast=True,
         regularizer=None,
         constraint=None,
-        aggregation="mean",
+        aggregation="none",
         name=None,
     ):
         """Add a weight variable to the layer.
@@ -499,10 +523,11 @@ class Layer(BackendLayer, Operation, KerasSaveable):
             constraint: Contrainst object to call on the variable after any
                 optimizer update, or string name of a built-in constraint.
                 Defaults to `None`.
-            aggregation: String, one of `'mean'`, `'sum'`,
-                `'only_first_replica'`. Annotates the variable with the type
-                of multi-replica aggregation to be used for this variable
-                when writing custom data parallel training loops.
+            aggregation: Optional string, one of `None`, `"none"`, `"mean"`,
+                `"sum"` or `"only_first_replica"`. Annotates the variable with
+                the type of multi-replica aggregation to be used for this
+                variable when writing custom data parallel training loops.
+                Defaults to `"none"`.
             name: String name of the variable. Useful for debugging purposes.
         """
         self._check_super_called()
@@ -841,7 +866,7 @@ class Layer(BackendLayer, Operation, KerasSaveable):
                 arg_name = list(call_spec.tensor_arguments_dict.keys())[0]
                 only_tensor_arg = call_spec.tensor_arguments_dict[arg_name]
                 mask = tree.map_structure(
-                    lambda x: getattr(x, "_keras_mask", None),
+                    backend.get_keras_mask,
                     only_tensor_arg,
                 )
                 kwargs["mask"] = mask
@@ -850,11 +875,15 @@ class Layer(BackendLayer, Operation, KerasSaveable):
                 expected_mask_arg_name = f"{k}_mask"
                 if expected_mask_arg_name in call_spec.argument_names:
                     if call_spec.arguments_dict[expected_mask_arg_name] is None:
-                        mask = tree.map_structure(
-                            lambda x: getattr(x, "_keras_mask", None), v
-                        )
+                        mask = tree.map_structure(backend.get_keras_mask, v)
                         kwargs[expected_mask_arg_name] = mask
 
+        # We need to cache the `previous_mask` before `__call__` because the
+        # mask might be removed during the call, such as `MultiHeadAttention`.
+        previous_mask = tree.map_structure(
+            backend.get_keras_mask, call_spec.first_arg
+        )
+
         ####################
         # 7. Call the layer.
         try:
@@ -874,7 +903,6 @@ class Layer(BackendLayer, Operation, KerasSaveable):
                 elif self.compute_dtype != self.variable_dtype:
                     # Enter a new scope if our dtypes are "mixed".
                     new_scope = backend.AutocastScope(self.compute_dtype)
-
                 if new_scope is not None:
                     with new_scope:
                         outputs = super().__call__(*args, **kwargs)
@@ -901,12 +929,9 @@ class Layer(BackendLayer, Operation, KerasSaveable):
                         if backend.is_tensor(output):
                             self.add_loss(self.activity_regularizer(output))
 
-            # Set masks on outputs,
-            # provided only the first positional input arg and its mask.
+            # Set `previous_mask` on outputs if available. It is provided only
+            # for the first positional input arg and its mask.
             # TODO: consider extending this to all args and kwargs.
-            previous_mask = tree.map_structure(
-                lambda x: getattr(x, "_keras_mask", None), call_spec.first_arg
-            )
             if self.supports_masking:
                 self._set_mask_metadata(
                     call_spec.first_arg, outputs, previous_mask
@@ -925,16 +950,7 @@ class Layer(BackendLayer, Operation, KerasSaveable):
         return outputs
 
     def call(self, *args, **kwargs):
-        raise NotImplementedError(
-            f"Layer {self.__class__.__name__} does not have a `call()` "
-            "method implemented."
-        )
-
-    def quantized_call(self, *args, **kwargs):
-        raise NotImplementedError(
-            f"Layer {self.__class__.__name__} does not have a "
-            "`quantized_call()` method implemented."
-        )
+        raise self._not_implemented_error(self.call)
 
     @traceback_utils.filter_traceback
     def stateless_call(
@@ -989,7 +1005,6 @@ class Layer(BackendLayer, Operation, KerasSaveable):
         ```
         """
         self._check_super_called()
-
         if not self.built:
             raise ValueError(
                 f"To call stateless_call, {self.__class__.__name__} must be "
@@ -1026,7 +1041,16 @@ class Layer(BackendLayer, Operation, KerasSaveable):
             state_mapping=mapping, collect_losses=return_losses
         ) as scope:
             if self.dtype_policy.quantization_mode is not None:
-                outputs = self.quantized_call(*args, **kwargs)
+                if self._remat_mode is not None:
+                    outputs = self.rematerialized_call(
+                        self.quantized_call, *args, **kwargs
+                    )(*args, **kwargs)
+                else:
+                    outputs = self.quantized_call(*args, **kwargs)
+            elif self._remat_mode is not None:
+                outputs = self.rematerialized_call(self.call, *args, **kwargs)(
+                    *args, **kwargs
+                )
             else:
                 outputs = self.call(*args, **kwargs)
             if return_losses:
@@ -1088,9 +1112,9 @@ class Layer(BackendLayer, Operation, KerasSaveable):
 
     @utils.default
     def compute_output_shape(self, *args, **kwargs):
-        raise NotImplementedError(
-            f"Layer {self.__class__.__name__} should implement "
-            "`def compute_output_shape(self, input_shape)`."
+        raise self._not_implemented_error(
+            self.compute_output_shape,
+            "Should implement `def compute_output_shape(self, input_shape)`.",
         )
 
     def add_loss(self, loss):
@@ -1139,7 +1163,10 @@ class Layer(BackendLayer, Operation, KerasSaveable):
         for variable in self.trainable_weights:
             if variable.regularizer is None:
                 continue
-            if backend.in_stateless_scope():
+            if backend.in_stateless_scope() and not in_symbolic_scope():
+                # If in symbolic scope, we might get `None` from
+                # `get_current_value` in `backend.compute_output_spec`. So we
+                # assign `variable` instead.
                 v = backend.get_stateless_scope().get_current_value(variable)
             else:
                 v = variable
@@ -1170,11 +1197,13 @@ class Layer(BackendLayer, Operation, KerasSaveable):
         for layer in self._layers:
             layer._clear_losses()
 
-    def quantize(self, mode):
-        raise NotImplementedError(
-            f"Layer {self.__class__.__name__} does not have a `quantize()` "
-            "method implemented."
-        )
+    # Quantization-related (int8 and float8) methods
+
+    def quantized_build(self, input_shape, mode):
+        raise self._not_implemented_error(self.quantized_build)
+
+    def quantize(self, mode, type_check=True):
+        raise self._not_implemented_error(self.quantize)
 
     def _check_quantize_args(self, mode, compute_dtype):
         if not self.built:
@@ -1203,6 +1232,53 @@ class Layer(BackendLayer, Operation, KerasSaveable):
                 "'mixed_float16' before calling `quantize()`."
             )
 
+    def quantized_call(self, *args, **kwargs):
+        current_remat_mode = get_current_remat_mode()
+
+        if (
+            current_remat_mode != self._remat_mode
+            and current_remat_mode is not None
+        ):
+            warnings.warn(
+                f"The RematScope at call time ({current_remat_mode}) differs "
+                f"the one set during layer initialization "
+                f"({self._remat_mode}). "
+                f"Restoring the correct rematerialization mode "
+                f"{self._remat_mode} for this layer."
+            )
+        if self.quantization_mode == "int8":
+            return self._int8_call(*args, **kwargs)
+        elif self.quantization_mode == "float8":
+            return self._float8_call(*args, **kwargs)
+        else:
+            raise self._quantization_mode_error(self.quantization_mode)
+
+    def _int8_call(self, *args, **kwargs):
+        raise self._not_implemented_error(self._int8_call)
+
+    def _float8_call(self, *args, **kwargs):
+        raise self._not_implemented_error(self._float8_call)
+
+    def _not_implemented_error(self, attr, msg=None):
+        if callable(attr):
+            attr_name = attr.__name__
+            attr_type = "method"
+        else:
+            attr_name = str(attr)
+            attr_type = "attribute"
+        msg = " " + msg if msg is not None else ""
+        return NotImplementedError(
+            f"Layer {self.__class__.__name__} does not have a `{attr_name}` "
+            f"{attr_type} implemented.{msg}"
+        )
+
+    def _quantization_mode_error(self, mode):
+        return NotImplementedError(
+            "Invalid quantization mode. Expected one of "
+            f"{dtype_policies.QUANTIZATION_MODES}. "
+            f"Received: quantization_mode={mode}"
+        )
+
     def save_own_variables(self, store):
         """Saves the state of the layer.
 
@@ -1277,9 +1353,12 @@ class Layer(BackendLayer, Operation, KerasSaveable):
             self._tracker.lock()
         self._post_untrack_variable(variable)
 
-    def add_metric(self):
+    def add_metric(self, *args, **kwargs):
         # Permanently disabled
-        raise NotImplementedError
+        raise NotImplementedError(
+            "Layer `add_metric()` method is deprecated"
+            " add your metric in `Model.compile(metrics=[...]).`"
+        )
 
     def count_params(self):
         """Count the total number of scalars composing the weights.
@@ -1376,8 +1455,7 @@ class Layer(BackendLayer, Operation, KerasSaveable):
 
     def __repr__(self):
         return (
-            f"<{self.__class__.__name__} "
-            f"name={self.name}, built={self.built}>"
+            f"<{self.__class__.__name__} name={self.name}, built={self.built}>"
         )
 
     def __str__(self):
@@ -1392,6 +1470,20 @@ class Layer(BackendLayer, Operation, KerasSaveable):
             value = self._tracker.track(value)
         return super().__setattr__(name, value)
 
+    def __delattr__(self, name):
+        obj = getattr(self, name)
+        if isinstance(obj, backend.Variable):
+            import gc
+
+            # It will take a short amount of time for the corresponding buffer
+            # to be actually removed from the device.
+            # https://stackoverflow.com/a/74631949
+            self._untrack_variable(obj)
+            super().__delattr__(name)
+            gc.collect()
+        else:
+            super().__delattr__(name)
+
     def _check_super_called(self):
         if getattr(self, "_lock", True):
             raise RuntimeError(
@@ -1402,9 +1494,19 @@ class Layer(BackendLayer, Operation, KerasSaveable):
 
     def _assert_input_compatibility(self, arg_0):
         if self.input_spec:
-            input_spec.assert_input_compatibility(
-                self.input_spec, arg_0, layer_name=self.name
-            )
+            try:
+                input_spec.assert_input_compatibility(
+                    self.input_spec, arg_0, layer_name=self.name
+                )
+            except SystemError:
+                if backend.backend() == "torch":
+                    # TODO: The torch backend failed the ONNX CI with the error:
+                    # SystemError: <method '__int__' of 'torch._C.TensorBase'
+                    # objects> returned a result with an exception set
+                    # As a workaround, we are skipping this for now.
+                    pass
+                else:
+                    raise
 
     def _get_call_context(self):
         """Returns currently active `CallContext`."""
@@ -1444,7 +1546,7 @@ class Layer(BackendLayer, Operation, KerasSaveable):
         flat_outputs = tree.flatten(outputs)
 
         mask_already_computed = all(
-            getattr(x, "_keras_mask", None) is not None for x in flat_outputs
+            backend.get_keras_mask(x) is not None for x in flat_outputs
         )
         if mask_already_computed:
             return
@@ -1455,18 +1557,14 @@ class Layer(BackendLayer, Operation, KerasSaveable):
 
         flat_masks = tree.flatten(output_masks)
         for tensor, mask in zip(flat_outputs, flat_masks):
-            if getattr(tensor, "_keras_mask", None) is None:
-                try:
-                    # Numpy backend does not support masking.
-                    if backend.backend() == "numpy":
-                        warnings.warn(
-                            "The NumPy backend does not support masking at this"
-                            "time. Masks will be ignored."
-                        )
-                    tensor._keras_mask = mask
-                except AttributeError:
-                    # It's a C type.
-                    pass
+            if backend.get_keras_mask(tensor) is None and mask is not None:
+                if backend.backend() == "numpy":
+                    warnings.warn(
+                        "The NumPy backend does not support masking at this"
+                        "time. Masks will be ignored."
+                    )
+                else:
+                    backend.set_keras_mask(tensor, mask)
 
     @python_utils.default
     def get_config(self):
@@ -1487,6 +1585,62 @@ class Layer(BackendLayer, Operation, KerasSaveable):
             self._parent_path = current_path()
         return backend.name_scope(self.name, caller=self)
 
+    def rematerialized_call(self, layer_call, *args, **kwargs):
+        """Enable rematerialization dynamically for layer's call method.
+
+        Args:
+            layer_call: The original `call` method of a layer.
+
+        Returns:
+            Rematerialized layer's `call` method.
+        """
+
+        def compute_size(x):
+            return (
+                math.prod([d or 1 for d in x.shape])
+                if isinstance(x, KerasTensor)
+                else 0
+            )
+
+        # Full rematerialization
+        if self._remat_mode.mode == "full":
+            return remat.remat(layer_call)
+
+        # Apply rematerialization to specific layers
+        elif self._remat_mode.mode == "list_of_layers" and (
+            self.name in self._remat_mode.layer_names
+        ):
+            return remat.remat(layer_call)
+
+        # Apply rematerialization based on output size threshold
+        elif self._remat_mode.mode == "larger_than":
+            output_spec = self.compute_output_spec(*args, **kwargs)
+            output_size = sum(
+                tree.flatten(tree.map_structure(compute_size, output_spec))
+            )
+            if (
+                output_size
+                and output_size > self._remat_mode.output_size_threshold
+            ):
+                return remat.remat(layer_call)
+        elif self._remat_mode.mode == "activations":
+            has_activation = (
+                hasattr(self, "activation") and self.activation is not None
+            )
+            if has_activation:
+
+                @functools.wraps(layer_call)
+                def rematerialized_activation_call_wrapper(*args, **kwargs):
+                    original_activation = self.activation
+                    self.activation = remat.remat(original_activation)
+                    try:
+                        return layer_call(*args, **kwargs)
+                    finally:
+                        self.activation = original_activation
+
+                return rematerialized_activation_call_wrapper
+        return layer_call
+
 
 def is_backend_tensor_or_symbolic(x, allow_none=False):
     if allow_none and x is None:
