@@ -1,3 +1,4 @@
+from keras.src import backend
 from keras.src import constraints
 from keras.src import initializers
 from keras.src import ops
@@ -85,7 +86,9 @@ class LayerNormalization(Layer):
         rms_scaling: If True, `center` and `scale` are ignored, and the
             inputs are scaled by `gamma` and the inverse square root
             of the square of all inputs. This is an approximate and faster
-            approach that avoids ever computing the mean of the input.
+            approach that avoids ever computing the mean of the input. Note that
+            this *isn't* equivalent to the computation that the
+            `keras.layers.RMSNormalization` layer performs.
         beta_initializer: Initializer for the beta weight. Defaults to zeros.
         gamma_initializer: Initializer for the gamma weight. Defaults to ones.
         beta_regularizer: Optional regularizer for the beta weight.
@@ -117,7 +120,7 @@ class LayerNormalization(Layer):
         gamma_regularizer=None,
         beta_constraint=None,
         gamma_constraint=None,
-        **kwargs
+        **kwargs,
     ):
         super().__init__(**kwargs)
         if isinstance(axis, (list, tuple)):
@@ -179,7 +182,6 @@ class LayerNormalization(Layer):
         self.built = True
 
     def call(self, inputs):
-        inputs = ops.cast(inputs, self.compute_dtype)
         # Compute the axes along which to reduce the mean / variance
         input_shape = inputs.shape
         ndims = len(input_shape)
@@ -199,11 +201,10 @@ class LayerNormalization(Layer):
                 return ops.reshape(v, broadcast_shape)
             return v
 
-        input_dtype = inputs.dtype
-        if input_dtype in ("float16", "bfloat16") and self.dtype == "float32":
-            # If mixed precision is used, cast inputs to float32 so that
-            # this is at least as numerically stable as the fused version.
-            inputs = ops.cast(inputs, "float32")
+        compute_dtype = backend.result_type(inputs.dtype, "float32")
+        # LN is prone to overflow with float16/bfloat16 inputs, so we upcast to
+        # float32 for the subsequent computations.
+        inputs = ops.cast(inputs, compute_dtype)
 
         if self.rms_scaling:
             # Calculate outputs with only variance and gamma if rms scaling
@@ -231,10 +232,21 @@ class LayerNormalization(Layer):
                 res = res + beta
 
             outputs = inputs * inv + res
-
-        return ops.cast(outputs, input_dtype)
+        return ops.cast(outputs, self.compute_dtype)
 
     def compute_output_shape(self, input_shape):
+        if isinstance(self.axis, int):
+            axes = [self.axis]
+        else:
+            axes = self.axis
+
+        for axis in axes:
+            if axis >= len(input_shape) or axis < -len(input_shape):
+                raise ValueError(
+                    f"Axis {axis} is out of bounds for "
+                    f"input shape {input_shape}. "
+                    f"Received: axis={self.axis}"
+                )
         return input_shape
 
     def get_config(self):
