@@ -372,8 +372,12 @@ def load_weights_from_hdf5_group(f, model):
                 f"Layer expects {len(symbolic_weights)} weight(s). Received "
                 f"{len(weight_values)} saved weight(s)"
             )
-        for ref_v, val in zip(symbolic_weights, weight_values):
-            ref_v.assign(val)
+        _set_weights(
+            layer,
+            symbolic_weights,
+            weight_values,
+            name=f"layer #{k} (named {layer.name})",
+        )
 
     if "top_level_model_weights" in f:
         symbolic_weights = list(
@@ -392,8 +396,58 @@ def load_weights_from_hdf5_group(f, model):
                 f"Model expects {len(symbolic_weights)} top-level weight(s). "
                 f"Received {len(weight_values)} saved top-level weight(s)"
             )
-        for ref_v, val in zip(symbolic_weights, weight_values):
-            ref_v.assign(val)
+        _set_weights(
+            model,
+            symbolic_weights,
+            weight_values,
+            name="top-level model",
+        )
+
+
+def _set_weights(
+    instance, symbolic_weights, weight_values, name, skip_mismatch=False
+):
+    """Safely set weights into a model or a layer.
+
+    Args:
+        instance: Model or layer instance,
+        symbolic_weights: symbolic tensors representing
+                        the weights of the variables to load,
+        weight_values: values of the weights to load,
+        skip_mismatch: Boolean, whether to skip loading of weights
+            where there is a mismatch in the shape of the weights,
+        name: name used to identify the group.
+
+    Raises:
+        ValueError: in case of mismatch between provided
+            model/layer and weights.
+    """
+    for i, weight_value in enumerate(weight_values):
+        expected_shape = symbolic_weights[i].shape
+        received_shape = weight_value.shape
+        if expected_shape != received_shape:
+            if skip_mismatch:
+                warnings.warn(
+                    f"Skipping loading weights for {name}"
+                    f"due to mismatch in shape for "
+                    f"weight {symbolic_weights[i].path}. "
+                    f"Weight expects shape {expected_shape}. "
+                    "Received saved weight "
+                    f"with shape {received_shape}",
+                    stacklevel=2,
+                )
+                continue
+            raise ValueError(
+                f"Shape mismatch in {name}"
+                f"for weight {symbolic_weights[i].path}. "
+                f"Weight expects shape {expected_shape}. "
+                "Received saved weight "
+                f"with shape {received_shape}"
+            )
+        symbolic_weights[i].assign(weight_value)
+
+    if hasattr(instance, "finalize_state") and symbolic_weights:
+        instance.finalize_state()
 
 
 def load_weights_from_hdf5_group_by_name(f, model, skip_mismatch=False):
@@ -456,33 +510,18 @@ def load_weights_from_hdf5_group_by_name(f, model, skip_mismatch=False):
                     f"Received {len(weight_values)} saved weight(s)"
                 )
             # Set values.
-            for i in range(len(weight_values)):
-                expected_shape = symbolic_weights[i].shape
-                received_shape = weight_values[i].shape
-                if expected_shape != received_shape:
-                    if skip_mismatch:
-                        warnings.warn(
-                            f"Skipping loading weights for layer #{k} (named "
-                            f"{layer.name}) due to mismatch in shape for "
-                            f"weight {symbolic_weights[i].path}. "
-                            f"Weight expects shape {expected_shape}. "
-                            "Received saved weight "
-                            f"with shape {received_shape}",
-                            stacklevel=2,
-                        )
-                        continue
-                    raise ValueError(
-                        f"Shape mismatch in layer #{k} (named {layer.name}) "
-                        f"for weight {symbolic_weights[i].path}. "
-                        f"Weight expects shape {expected_shape}. "
-                        "Received saved weight "
-                        f"with shape {received_shape}"
-                    )
-                else:
-                    symbolic_weights[i].assign(weight_values[i])
+            _set_weights(
+                layer,
+                symbolic_weights,
+                weight_values,
+                skip_mismatch=skip_mismatch,
+                name=f"layer #{k} (named {layer.name})",
+            )
 
     if "top_level_model_weights" in f:
-        symbolic_weights = model.trainable_weights + model.non_trainable_weights
+        symbolic_weights = (
+            model._trainable_variables + model._non_trainable_variables
+        )
         weight_values = load_subset_weights_from_hdf5_group(
             f["top_level_model_weights"]
         )
@@ -505,30 +544,13 @@ def load_weights_from_hdf5_group_by_name(f, model, skip_mismatch=False):
                     f"Received {len(weight_values)} saved top-level weight(s)"
                 )
         else:
-            for i in range(len(weight_values)):
-                expected_shape = symbolic_weights[i].shape
-                received_shape = weight_values[i].shape
-                if expected_shape != received_shape:
-                    if skip_mismatch:
-                        warnings.warn(
-                            "Skipping loading top-level weight for model due "
-                            "to mismatch in shape for "
-                            f"weight {symbolic_weights[i].path}. "
-                            f"Weight expects shape {expected_shape}. "
-                            "Received saved weight "
-                            f"with shape {received_shape}",
-                            stacklevel=2,
-                        )
-                    else:
-                        raise ValueError(
-                            "Shape mismatch in model for top-level weight "
-                            f"{symbolic_weights[i].path}. "
-                            f"Weight expects shape {expected_shape}. "
-                            "Received saved weight "
-                            f"with shape {received_shape}"
-                        )
-                else:
-                    symbolic_weights[i].assign(weight_values[i])
+            _set_weights(
+                model,
+                symbolic_weights,
+                weight_values,
+                skip_mismatch=skip_mismatch,
+                name="top-level model",
+            )
 
 
 def load_subset_weights_from_hdf5_group(f):
