@@ -240,7 +240,7 @@ class Resize(Operation):
         self.data_format = backend.standardize_data_format(data_format)
 
     def call(self, images):
-        return backend.image.resize(
+        return _resize(
             images,
             self.size,
             interpolation=self.interpolation,
@@ -362,7 +362,7 @@ def resize(
             fill_mode=fill_mode,
             fill_value=fill_value,
         ).symbolic_call(images)
-    return backend.image.resize(
+    return _resize(
         images,
         size,
         interpolation=interpolation,
@@ -375,6 +375,37 @@ def resize(
     )
 
 
+def _resize(
+    images,
+    size,
+    interpolation="bilinear",
+    antialias=False,
+    crop_to_aspect_ratio=False,
+    pad_to_aspect_ratio=False,
+    fill_mode="constant",
+    fill_value=0.0,
+    data_format=None,
+):
+    resized = backend.image.resize(
+        images,
+        size,
+        interpolation=interpolation,
+        antialias=antialias,
+        crop_to_aspect_ratio=crop_to_aspect_ratio,
+        data_format=data_format,
+        pad_to_aspect_ratio=pad_to_aspect_ratio,
+        fill_mode=fill_mode,
+        fill_value=fill_value,
+    )
+    if resized.dtype == images.dtype:
+        # Only `torch` backend will cast result to original dtype with
+        # correct rounding and without dtype overflow
+        return resized
+    if backend.is_int_dtype(images.dtype):
+        resized = ops.round(resized)
+    return ops.saturate_cast(resized, images.dtype)
+
+
 class AffineTransform(Operation):
     def __init__(
         self,
@@ -583,7 +614,7 @@ def extract_patches(
 
     Args:
         images: Input image or batch of images. Must be 3D or 4D.
-        size: Patch size int or tuple (patch_height, patch_widht)
+        size: Patch size int or tuple (patch_height, patch_width)
         strides: strides along height and width. If not specified, or
             if `None`, it defaults to the same value as `size`.
         dilation_rate: This is the input stride, specifying how far two
@@ -947,8 +978,7 @@ def _pad_images(
         )
     if left_padding < 0:
         raise ValueError(
-            "left_padding must be >= 0. "
-            f"Received: left_padding={left_padding}"
+            f"left_padding must be >= 0. Received: left_padding={left_padding}"
         )
     if right_padding < 0:
         raise ValueError(
@@ -1167,8 +1197,7 @@ def _crop_images(
 
     if top_cropping < 0:
         raise ValueError(
-            "top_cropping must be >= 0. "
-            f"Received: top_cropping={top_cropping}"
+            f"top_cropping must be >= 0. Received: top_cropping={top_cropping}"
         )
     if target_height < 0:
         raise ValueError(
@@ -1182,8 +1211,7 @@ def _crop_images(
         )
     if target_width < 0:
         raise ValueError(
-            "target_width must be >= 0. "
-            f"Received: target_width={target_width}"
+            f"target_width must be >= 0. Received: target_width={target_width}"
         )
 
     # Compute start_indices and shape
@@ -1202,3 +1230,231 @@ def _crop_images(
 
     cropped_images = ops.slice(images, start_indices, shape)
     return cropped_images
+
+
+class PerspectiveTransform(Operation):
+    def __init__(
+        self,
+        interpolation="bilinear",
+        fill_value=0,
+        data_format=None,
+    ):
+        super().__init__()
+        self.interpolation = interpolation
+        self.fill_value = fill_value
+        self.data_format = backend.standardize_data_format(data_format)
+
+    def call(self, images, start_points, end_points):
+        return backend.image.perspective_transform(
+            images,
+            start_points,
+            end_points,
+            interpolation=self.interpolation,
+            fill_value=self.fill_value,
+            data_format=self.data_format,
+        )
+
+    def compute_output_spec(self, images, start_points, end_points):
+        if len(images.shape) not in (3, 4):
+            raise ValueError(
+                "Invalid images rank: expected rank 3 (single image) "
+                "or rank 4 (batch of images). Received input with shape: "
+                f"images.shape={images.shape}"
+            )
+        if start_points.shape[-2:] != (4, 2) or start_points.ndim not in (2, 3):
+            raise ValueError(
+                "Invalid start_points shape: expected (4,2) for a single image"
+                f" or (N,4,2) for a batch. Received shape: {start_points.shape}"
+            )
+        if end_points.shape[-2:] != (4, 2) or end_points.ndim not in (2, 3):
+            raise ValueError(
+                "Invalid end_points shape: expected (4,2) for a single image"
+                f" or (N,4,2) for a batch. Received shape: {end_points.shape}"
+            )
+        if start_points.shape != end_points.shape:
+            raise ValueError(
+                "start_points and end_points must have the same shape."
+                f" Received start_points.shape={start_points.shape}, "
+                f"end_points.shape={end_points.shape}"
+            )
+        return KerasTensor(images.shape, dtype=images.dtype)
+
+
+@keras_export("keras.ops.image.perspective_transform")
+def perspective_transform(
+    images,
+    start_points,
+    end_points,
+    interpolation="bilinear",
+    fill_value=0,
+    data_format=None,
+):
+    """Applies a perspective transformation to the image(s).
+
+    Args:
+        images: Input image or batch of images. Must be 3D or 4D.
+        start_points: A tensor of shape `(N, 4, 2)` or `(4, 2)`,
+            representing the source points in the original image
+            that define the transformation.
+        end_points: A tensor of shape `(N, 4, 2)` or `(4, 2)`,
+            representing the target points in the output image
+            after transformation.
+        interpolation: Interpolation method. Available methods are `"nearest"`,
+            and `"bilinear"`. Defaults to `"bilinear"`.
+        fill_value: Value used for points outside the boundaries of the input if
+            extrapolation is needed. Defaults to `0`.
+        data_format: A string specifying the data format of the input tensor.
+            It can be either `"channels_last"` or `"channels_first"`.
+            `"channels_last"` corresponds to inputs with shape
+            `(batch, height, width, channels)`, while `"channels_first"`
+            corresponds to inputs with shape `(batch, channels, height, width)`.
+            If not specified, the value will default to
+            `keras.config.image_data_format`.
+
+    Returns:
+        Applied perspective transform image or batch of images.
+
+    Examples:
+
+    >>> x = np.random.random((2, 64, 80, 3))  # batch of 2 RGB images
+    >>> start_points = np.array(
+    ...     [
+    ...         [[0, 0], [0, 64], [80, 0], [80, 64]],
+    ...         [[0, 0], [0, 64], [80, 0], [80, 64]],
+    ...     ]
+    ... )
+    >>> end_points = np.array(
+    ...     [
+    ...         [[3, 5], [7, 64], [76, -10], [84, 61]],
+    ...         [[8, 10], [10, 61], [65, 3], [88, 43]],
+    ...     ]
+    ... )
+    >>> y = keras.ops.image.perspective_transform(x, start_points, end_points)
+    >>> y.shape
+    (2, 64, 80, 3)
+
+    >>> x = np.random.random((64, 80, 3))  # single RGB image
+    >>> start_points = np.array([[0, 0], [0, 64], [80, 0], [80, 64]])
+    >>> end_points = np.array([[3, 5], [7, 64], [76, -10], [84, 61]])
+    >>> y = keras.ops.image.perspective_transform(x, start_points, end_points)
+    >>> y.shape
+    (64, 80, 3)
+
+    >>> x = np.random.random((2, 3, 64, 80))  # batch of 2 RGB images
+    >>> start_points = np.array(
+    ...     [
+    ...         [[0, 0], [0, 64], [80, 0], [80, 64]],
+    ...         [[0, 0], [0, 64], [80, 0], [80, 64]],
+    ...     ]
+    ... )
+    >>> end_points = np.array(
+    ...     [
+    ...         [[3, 5], [7, 64], [76, -10], [84, 61]],
+    ...         [[8, 10], [10, 61], [65, 3], [88, 43]],
+    ...     ]
+    ... )
+    >>> y = keras.ops.image.perspective_transform(
+    ...     x, start_points, end_points, data_format="channels_first"
+    ... )
+    >>> y.shape
+    (2, 3, 64, 80)
+    """
+    if any_symbolic_tensors((images, start_points, end_points)):
+        return PerspectiveTransform(
+            interpolation=interpolation,
+            fill_value=fill_value,
+            data_format=data_format,
+        ).symbolic_call(images, start_points, end_points)
+    return backend.image.perspective_transform(
+        images,
+        start_points,
+        end_points,
+        interpolation=interpolation,
+        fill_value=fill_value,
+        data_format=data_format,
+    )
+
+
+class GaussianBlur(Operation):
+    def __init__(
+        self,
+        kernel_size=(3, 3),
+        sigma=(1.0, 1.0),
+        data_format=None,
+    ):
+        super().__init__()
+        self.kernel_size = kernel_size
+        self.sigma = sigma
+        self.data_format = backend.standardize_data_format(data_format)
+
+    def call(self, images):
+        return backend.image.gaussian_blur(
+            images,
+            kernel_size=self.kernel_size,
+            sigma=self.sigma,
+            data_format=self.data_format,
+        )
+
+    def compute_output_spec(self, images):
+        if len(images.shape) not in (3, 4):
+            raise ValueError(
+                "Invalid images rank: expected rank 3 (single image) "
+                "or rank 4 (batch of images). Received input with shape: "
+                f"images.shape={images.shape}"
+            )
+        return KerasTensor(images.shape, dtype=images.dtype)
+
+
+@keras_export("keras.ops.image.gaussian_blur")
+def gaussian_blur(
+    images, kernel_size=(3, 3), sigma=(1.0, 1.0), data_format=None
+):
+    """Applies a Gaussian blur to the image(s).
+
+    Args:
+        images: Input image or batch of images. Must be 3D or 4D.
+        kernel_size: A tuple of two integers, specifying the height and width
+            of the Gaussian kernel.
+        sigma: A tuple of two floats, specifying the standard deviation of
+            the Gaussian kernel along height and width.
+        data_format: A string specifying the data format of the input tensor.
+            It can be either `"channels_last"` or `"channels_first"`.
+            `"channels_last"` corresponds to inputs with shape
+            `(batch, height, width, channels)`, while `"channels_first"`
+            corresponds to inputs with shape `(batch, channels, height, width)`.
+            If not specified, the value will default to
+            `keras.config.image_data_format`.
+
+    Returns:
+        Blurred image or batch of images.
+
+    Examples:
+
+    >>> x = np.random.random((2, 64, 80, 3))  # batch of 2 RGB images
+    >>> y = keras.ops.image.gaussian_blur(x)
+    >>> y.shape
+    (2, 64, 80, 3)
+
+    >>> x = np.random.random((64, 80, 3))  # single RGB image
+    >>> y = keras.ops.image.gaussian_blur(x)
+    >>> y.shape
+    (64, 80, 3)
+
+    >>> x = np.random.random((2, 3, 64, 80))  # batch of 2 RGB images
+    >>> y = keras.ops.image.gaussian_blur(
+    ...     x, data_format="channels_first")
+    >>> y.shape
+    (2, 3, 64, 80)
+    """
+    if any_symbolic_tensors((images,)):
+        return GaussianBlur(
+            kernel_size=kernel_size,
+            sigma=sigma,
+            data_format=data_format,
+        ).symbolic_call(images)
+    return backend.image.gaussian_blur(
+        images,
+        kernel_size=kernel_size,
+        sigma=sigma,
+        data_format=data_format,
+    )
