@@ -47,7 +47,6 @@ def _cholesky(x):
 
 
 class Det(Operation):
-
     def __init__(self):
         super().__init__()
 
@@ -68,7 +67,7 @@ def det(x):
         x: Input tensor of shape `(..., M, M)`.
 
     Returns:
-        A tensor of shape `(...,)` represeting the determinant of `x`.
+        A tensor of shape `(...,)` representing the determinant of `x`.
 
     """
     if any_symbolic_tensors((x,)):
@@ -84,7 +83,6 @@ def _det(x):
 
 
 class Eig(Operation):
-
     def __init__(self):
         super().__init__()
 
@@ -124,7 +122,6 @@ def _eig(x):
 
 
 class Eigh(Operation):
-
     def __init__(self):
         super().__init__()
 
@@ -165,7 +162,6 @@ def _eigh(x):
 
 
 class Inv(Operation):
-
     def __init__(self):
         super().__init__()
 
@@ -202,7 +198,6 @@ def _inv(x):
 
 
 class LuFactor(Operation):
-
     def __init__(self):
         super().__init__()
 
@@ -445,7 +440,6 @@ def qr(x, mode="reduced"):
 
 
 class Solve(Operation):
-
     def __init__(self):
         super().__init__()
 
@@ -466,7 +460,7 @@ def solve(a, b):
 
     Args:
         a: A tensor of shape `(..., M, M)` representing the coefficients matrix.
-        b: A tensor of shape `(..., M)` or `(..., M, N)` represeting the
+        b: A tensor of shape `(..., M)` or `(..., M, N)` representing the
         right-hand side or "dependent variable" matrix.
 
     Returns:
@@ -490,7 +484,6 @@ def _solve(a, b):
 
 
 class SolveTriangular(Operation):
-
     def __init__(self, lower=False):
         super().__init__()
         self.lower = lower
@@ -514,7 +507,7 @@ def solve_triangular(a, b, lower=False):
 
     Args:
         a: A tensor of shape `(..., M, M)` representing the coefficients matrix.
-        b: A tensor of shape `(..., M)` or `(..., M, N)` represeting the
+        b: A tensor of shape `(..., M)` or `(..., M, N)` representing the
         right-hand side or "dependent variable" matrix.
 
     Returns:
@@ -538,7 +531,6 @@ def _solve_triangular(a, b, lower=False):
 
 
 class SVD(Operation):
-
     def __init__(self, full_matrices=True, compute_uv=True):
         super().__init__()
         self.full_matrices = full_matrices
@@ -604,12 +596,11 @@ class Lstsq(Operation):
     def compute_output_spec(self, a, b):
         if len(a.shape) != 2:
             raise ValueError(
-                "Expected a to have rank 2. " f"Received: a.shape={a.shape}"
+                f"Expected a to have rank 2. Received: a.shape={a.shape}"
             )
         if len(b.shape) not in (1, 2):
             raise ValueError(
-                "Expected b to have rank 1 or 2. "
-                f"Received: b.shape={b.shape}"
+                f"Expected b to have rank 1 or 2. Received: b.shape={b.shape}"
             )
         m, n = a.shape
         if b.shape[0] != m:
@@ -674,8 +665,7 @@ def _assert_1d(*arrays):
     for a in arrays:
         if a.ndim < 1:
             raise ValueError(
-                "Expected input to have rank >= 1. "
-                "Received scalar input {a}."
+                "Expected input to have rank >= 1. Received scalar input {a}."
             )
 
 
